// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of scatter
//
// scatter is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// scatter is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the
//       Free Software Foundation, Inc.
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include  "MZN.h"

#include <mathconst/mathconst.h>

#include "saotrace/io/fits/fits.h"

namespace FITS = SAOTrace::IO::FITS;

MZN::MZN( const string& id, const string& filename, double zmin, double zmax )
    throw ( Exception ) : MZ(id, filename, zmin, zmax ) {

  TraceFct tf("MZN::MZN");

  vector <FITS::ColumnBase*> columns;

  FITS::Column<double> esa  ( "ESA"  , _esa,        columns );
  FITS::Column<double> theta( "THETA", _theta.data, columns );

  _nrows = FITS::read_table_columns( _file, columns );

  _theta.ncols = theta.extent[0];

}

bool
MZN::lookUpScatterAngle( double energy, double sin_alpha, double rannum, bool inplane, double& theta )
{

#ifdef DEBUG
  TraceFct tf( "MZN::lookUpScatterAngle" );
#endif // DEBUG

  double esa = energy * sin_alpha;

  double *esa_lo, *esa_hi;

  bool is_extrapolated;

  if ( NULL != bnd_bsearch( &esa, &_esa[0], _nrows, sizeof(_esa[0]),
			    (const void **) &esa_lo, (const void **) &esa_hi, esacmp ) ) {            // Exact match!!!

    is_extrapolated = false;

  } else if ( NULL == esa_lo ) { // Less than all _esa

    esa_lo  = esa_hi;
    is_extrapolated = true;

  } else if ( NULL == esa_hi ) { // Greater than all _esa

    esa_hi  = esa_lo;
    is_extrapolated = true;

  } else {                       // Within _esa

    is_extrapolated = false;

  }

  long idx_lo = esa_lo - &_esa[0];
  long idx_hi = esa_hi - &_esa[0];

#ifdef DEBUG
  if ( dbflag( "esa" ) )
  {
    tf.message(
	        "sin_alpha:\t%g\n"
	        "esa_lo:\t%g\n"
		"esa:\t%g\n"
		"esa_hi:\t%g\n"
		"esaidx_lo:\t%ld\n"
		"esaidx_hi:\t%ld\n"
		, sin_alpha, *esa_lo, esa, *esa_hi, idx_lo, idx_hi );
  }
#endif

  if ( idx_lo != idx_hi )
  {
      double theta_lo = lookUpTheta( idx_lo, rannum );
      double theta_hi = lookUpTheta( idx_hi, rannum );
      theta = interpolate( esa, *esa_lo, theta_lo, *esa_hi, theta_hi );
  }
  else
  {
      theta = lookUpTheta( idx_lo, rannum );
  }

  // out-of-plane scattering? randomize the sign of the scattering angle
  // to get a symmetric distribution
  if ( ! inplane )
  {
      theta *= sin_alpha;
      if ( ranf() > 0.5 )
	  theta *= -1;
  }


#ifdef DEBUG
    if ( dbflag( "theta" ) )
      tf.message(  "theta:\t%g\n", theta );
#endif

  // convert theta to radians and flip its direction.  MZ::scatter
  // expects positive scattering angles in the backscatter direction.
  theta *= - M_PI  / ( 3600 * 180 );

  return is_extrapolated;
}

double
MZN::lookUpTheta( const long idx, double rannum ) {


  long ndist = _theta.ncols;

  double didx = rannum * (double) ndist;

  long jdx = (long) didx;

  if ( jdx >= ndist )
    jdx = ndist - 1;

  double f2 = didx - (double) jdx;
  double f1 = 1 - f2;

  double theta = f1 * _theta( idx, jdx ) + f2 * _theta(idx, jdx + 1);

#ifdef DEBUG
  TraceFct tf( "MZN::lookUpTheta" );

  if ( dbflag( "idx" ) )
    tf.message( "idx:\t%d\n"
		"jidx:\t%d\n"
		"rannum:\t%g\n"
		"f1:\t%g\n"
		"f2:\t%g\n"
		"theta_lookup:\t%g\n", idx, jdx, rannum, f1, f2, theta );

#endif // DEBUG

  return theta;
}



