// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of scatter
//
// scatter is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// scatter is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the
//       Free Software Foundation, Inc.
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <iomanip>
#include <iostream>
#include <string>

#include <suplib/debug.h>
#include <tracefctxx/TraceFct.h>
#include <bpipexx/bpipexx.h>
#include <Exception/Exception.h>

#include "MirrorDatabase.h"
#include "MZ.h"

#include "Photon.h"
#include "Random.h"
#include "ScatterDatabase.h"

#include "process.h"

using namespace SAOTrace;

void process( cxcparamx::simple& pf ) throw ( Exception ) {

  TraceFct tf( "process" );

  try {
    bool dump_offzone      = dbflag( "offzone" );
    bool dump_offzone2     = dbflag( "offzone2" );

    bool in_plane          = pf.getb( "in_plane" );
    bool out_of_plane      = pf.getb( "out_of_plane" );
    int  iter_limit        = pf.geti( "scat_iter_lim" );

    double theta_def       = ARCSEC2RAD(pf.getd( "theta" ));
    double phi_def         = ARCSEC2RAD(pf.getd( "phi" ));
    double max_cum_prob    = pf.getd( "max_cumulative_probability" );
    double min_cum_prob    = pf.getd( "min_cumulative_probability" );

    // Initialize the mirror position and scatter databases...
    MirrorDatabase mirrordb( pf.getstr( "mirror_db" ),
			     pf.geti( "surf_no" ),
			     pf.geti( "shell" ) ) ;

    ScatterDatabase scatterdb( pf.getstr( "scatter_db" ),
			       pf.geti( "surf_no" ),
			       pf.geti( "shell" ),
			       mirrordb.z0( ) );

    // Initialize I/O...
    BPipeX::Simple bpipe( pf.getstr( "input" ), pf.getstr( "output" ) );
    Scatter::Photon photon( bpipe );

    //------------------------------------------------------------------

    unsigned int surf_missed_mask = 0;

    // don't need srfno, but if it's there, use it.
    try {

	unsigned int* srfno 	 = bpipe.hdrf_data<unsigned int>( "srfno" );
	surf_missed_mask = 1 << (*srfno - 1);

    } catch ( ... ) { };

    // Initialize random number generator...
    Random randy( pf.geti( "seed1" ), pf.geti( "seed2" ), pf.geti( "block" ) );

    // Read photons...
    unsigned long n_offzone         = 0;
    unsigned long n_ip_extrapolated = 0;
    unsigned long n_op_extrapolated = 0;
    unsigned long n_ghosts          = 0;
    unsigned long n_surface_scat    = 0;
    unsigned long n_backscattered   = 0;

    while ( bpipe.read_dpkts( ) ) {

      // If it hit the last surface let's try to scatter it...
	if ( ! ( surf_missed_mask & photon.surf_missed() ) ) {

	    // Look at Z position of photon and select an appropriate
	    // mirror zone...
	    MZ *zone = scatterdb.findZone( photon.position().z );

	// If this photon fell outside all mirror zones...
	if ( ! zone ) {

	  if ( dump_offzone2 ) {

	    double diff;
	    if ( photon.position().z < scatterdb.zones( )[0]->zmin() )
		diff = scatterdb.zones( )[0]->zmin() - photon.position().z;
	    else
	      diff = photon.position().z - scatterdb.zones( )[scatterdb.zones( ).size( )-1]->zmax();

	    tf.message( "%u:  not in zone.  z=%f (%f,%f) diff = %f\n", photon.id(), photon.position().z,
			scatterdb.zones( )[0]->zmin(),
			scatterdb.zones( )[scatterdb.zones( ).size( )-1]->zmax(), diff );

	  }

	  n_offzone++;

	  if ( dump_offzone )
	      tf.message( "%u:  not in zone.  z = %f\n", photon.id(), photon.position().z );

	  continue;

	}

	// YOU HAVEN'T IMPLEMENTED THE skip ray if scattering is not
	// requested LOGIC!!!  is this option necessary?
	// skip ray if scattering is not requested
	// if ( ! zone->flag || NULL == zone->scat_table )
	//   continue;

	// Finally, scatter the photon and, if it has scattered into
	// the surface or backscattered, absorb it.

	int status = zone->scatter( in_plane, out_of_plane, iter_limit,
				    theta_def, phi_def, min_cum_prob, max_cum_prob,
				    photon.id(),
				    photon.energy(),
				    photon.surface_norm(),
				    *photon.p_direction(),
				    *photon.p_theta(),
				    *photon.p_phi() );

	if ( status & MZ::SCATTERED_INTO_SURFACE )
	{
          tf.message( "(%u): scattered into surface\n", photon.id() );
	  n_surface_scat++;
	  continue;
	}

	if ( status & MZ::BACKSCATTERED )
	{
	  tf.message( "(%u): backscattered\n", photon.id() );
	  n_backscattered++;
	  continue;
	}

	if ( status & MZ::IP_EXTRAPOLATED )
	{
	  n_ip_extrapolated++;

	  if ( dbflag( "extrapolated" ) )
	    tf.message( "(%u): in-plane extrapolated\n", photon.id() );
	}

	if ( status & MZ::OP_EXTRAPOLATED )
	{
	  n_op_extrapolated++;

	  if ( dbflag( "extrapolated" ) )
	    tf.message( "(%u): out-of-plane extrapolated\n", photon.id() );
	}

      // Else it missed the last surface and we should count it as a ghost...
      } else {

	n_ghosts++;

      }

      // Write photons...
      bpipe.write_dpkts( );

    }

    tf.message( "summary:  in-plane extrapolated:  %d\n", n_ip_extrapolated );
    tf.message( "summary:  out-of-plane extrapolated:  %d\n", n_op_extrapolated );
    tf.message( "summary:  off zone:  %d\n", n_offzone );
    tf.message( "summary:  scattered into surface:  %d\n", n_surface_scat );
    tf.message( "summary:  backscattered:  %d\n", n_backscattered );
    tf.message( "summary:  ghosts:  %d\n", n_ghosts );

  } catch ( Exception& E ) {

    throw( E );

  } catch ( exception& e ) {

    throw( Exception( e.what( ) ) );

  }
}
